﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.AnalysisServices;
using System.Data;
using System.Data.SqlClient;
using System.Data.OleDb;

namespace AmoTabular {
    /// <summary>
    /// 
    ///     This class is a very simple shell around AMO that simplifies the management of a Tabular
    ///     database. Its purpose is to make it easier to create a Tabular database by handling the
    ///     complexity of the translation of the Tabular concepts of tables, calculated columns and
    ///     measures in Multidimensional items like dimensions, measure groups, cubes and data source
    ///     views.
    ///     
    ///     You can find a full explanation of the code in the book
    ///     
    ///     Most of the code is a direct porting of the Amo2Tabular demo project provided by
    ///     Microsoft as a part of the SQL 2012 demo
    /// 
    /// </summary>
    public class TabularProject {
        const string rowNumberColumnName = "InternalRowNumber";
        const string cubeName = "Model";

        public Server Server;
        public Database Database;


        public TabularProject () {
            Server = new Server ();
        }

        /// <summary>
        /// 
        ///     Creates the database, destroying it if it already exists
        /// 
        /// </summary>
        /// <param name="databaseName"></param>
        public void CreateDatabase (string databaseName) {
            if (Server.Databases.ContainsName (databaseName)) {
                Server.Databases[databaseName].Drop (DropOptions.IgnoreFailures);
            }
            Database = new Database (databaseName);
            Database.StorageEngineUsed = StorageEngineUsed.InMemory;
            Server.Databases.Add (Database);
            Database.Update ();
        }

        /// <summary>
        /// 
        ///     Creates a new connection
        ///     
        ///     In reality, it creates a DataSource which, in Tabular, is called a Connection.
        ///     For demo purposes, this method works with SQL Server only.
        ///     Please note that a Multidimensional model needs a DataSourceView and a cube to work,
        ///     because a cube can be created only if a DataSourceView and a DataSourceView can be
        ///     created only if a DataSource exists, here se create both items after the creation of
        ///     the first connection.
        ///     The DataSourceView will be linked to the first DataSource (i.e. connection) created but
        ///     this does not matter, because whenever a table is added to the model, it is linked to
        ///     its DataSource, which does not need to be the same DataSOurce of the DataSOurceView
        /// 
        /// </summary>
        /// <param name="Database"></param>
        /// <param name="serverName"></param>
        /// <param name="databaseName"></param>
        public void AddSqlServerConnection (string serverName, string databaseName) {
            StringBuilder connString = new StringBuilder ();
            connString.AppendLine (String.Format ("Data Source={0};", serverName));
            connString.AppendLine (String.Format ("Initial Catalog={0};", databaseName));
            connString.AppendLine (String.Format ("Persist Security Info=false;"));
            connString.AppendLine (String.Format ("Integrated Security=SSPI;"));
            connString.AppendLine (String.Format ("Provider=SQLNCLI11;"));
            DataSource newDataSource = Database.DataSources.AddNew (databaseName);
            newDataSource.ConnectionString = connString.ToString ();
            newDataSource.ImpersonationInfo = new ImpersonationInfo (ImpersonationMode.ImpersonateServiceAccount);
            newDataSource.Update ();

            //
            //  If this is the first connection, create the data source view, the model cube and
            //  the empty MDX script
            //
            if (Database.DataSources.Count == 1) {
                //
                //  Create the new data source view as an empty DSV
                //
                DataSourceView newDataSourceView = Database.DataSourceViews.AddNew ("Sandbox");
                newDataSourceView.DataSourceID = newDataSource.ID;
                newDataSourceView.Schema = new DataSet ();
                newDataSourceView.Update ();

                //
                //  Create Model Cube
                //
                Cube newModelCube = Database.Cubes.Add (cubeName);
                newModelCube.Source = new DataSourceViewBinding (newDataSourceView.ID);
                newModelCube.StorageMode = StorageMode.InMemory;
                newModelCube.Language = Database.Language;
                newModelCube.Collation = Database.Collation;

                //
                //  Create initial MdxScript
                //
                MdxScript mdxScript = newModelCube.MdxScripts.Add ("MdxScript", "MdxScript");
                StringBuilder initialCommand = new StringBuilder ();
                initialCommand.AppendLine ("CALCULATE;");
                initialCommand.AppendLine ("CREATE MEMBER CURRENTCUBE.Measures.[__No measures defined] AS 1;");
                initialCommand.AppendLine ("ALTER CUBE CURRENTCUBE UPDATE DIMENSION Measures, Default_Member = [__No measures defined];");
                mdxScript.Commands.Add (new Command (initialCommand.ToString ()));
            }
        }

        /// <summary>
        /// 
        ///     This method demostrates how to create a table in Tabular.
        ///     It works only with a connection with SQL Server due to the way we get the
        ///     metadata from the table to fill the data source view schema
        /// 
        /// </summary>
        /// <param name="dataSourceName"></param>
        /// <param name="tableName"></param>
        /// <param name="tableQuery"></param>
        public void AddTable (string dataSourceName, string tableName, string tableQuery) {
            DataSourceView dataSourceView = Database.DataSourceViews["Sandbox"];
            DataSource dataSource = Database.DataSources[dataSourceName];
            Cube modelCube = Database.Cubes[0];

            //
            //  Retrieves table metadata and add them to the data source view
            //
            DataTable dataTable = new DataTable (tableName);
            string metaDataConnString = dataSource.ConnectionString.Replace ("Provider=SQLNCLI11;", "");
            using (SqlConnection conn = new SqlConnection (metaDataConnString)) {
                SqlDataAdapter dataAdapter = new SqlDataAdapter (tableQuery, conn);
                dataAdapter.FillSchema (dataTable, SchemaType.Source);
            }
            //
            //  Sets the DataSourceID of the table before adding it to the DataSourceView
            //  because, otherwise, SSAS will search for the table in the DataSourceView 
            //  DataSource, which might not be the correct one. In Tabular you have a single
            //  DataSourceView with many tables from different DataSources, while in Multidimensional
            //  you have many data source views, one for each data source.
            //
            dataTable.ExtendedProperties.Add ("DataSourceID", dataSource.ID);
            dataSourceView.Schema.Tables.Add (dataTable);

            //
            //  Create Dimension
            //
            Dimension newDimension = Database.Dimensions.AddNew (tableName);
            newDimension.Source = new DataSourceViewBinding (dataSourceView.ID);
            newDimension.StorageMode = DimensionStorageMode.InMemory;
            newDimension.UnknownMember = UnknownMemberBehavior.AutomaticNull;
            newDimension.UnknownMemberName = "Unknown";
            newDimension.ErrorConfiguration = new ErrorConfiguration ();
            newDimension.ErrorConfiguration.KeyNotFound = ErrorOption.IgnoreError;
            newDimension.ErrorConfiguration.KeyDuplicate = ErrorOption.ReportAndStop;
            newDimension.ErrorConfiguration.NullKeyNotAllowed = ErrorOption.ReportAndStop;
            newDimension.Language = Database.Language;
            newDimension.Collation = Database.Collation;
            newDimension.ProactiveCaching = new ProactiveCaching ();
            TimeSpan defaultProactiveChachingTimeSpan = new TimeSpan (0, 0, -1);
            newDimension.ProactiveCaching.SilenceInterval = defaultProactiveChachingTimeSpan;
            newDimension.ProactiveCaching.Latency = defaultProactiveChachingTimeSpan;
            newDimension.ProactiveCaching.SilenceOverrideInterval = defaultProactiveChachingTimeSpan;
            newDimension.ProactiveCaching.ForceRebuildInterval = defaultProactiveChachingTimeSpan;
            newDimension.ProactiveCaching.Source = new ProactiveCachingInheritedBinding ();

            //
            //  Manualy add a "RowNumber" attribute as the key attribute of the dimension, until a primary key is defined
            //  "RowNumber" a required column for a tabular model and has to be of type AMO.AttributeType.RowNumber and binding AMO.RowNumberBinding
            //  The name of the "RowNumber" attribute can be any name, as long as type and binding are correctly set
            //  By default the MS client tools set the column name and column ID of the RowNumber attribute to "RowNumber"
            //
            DimensionAttribute rowNumber = newDimension.Attributes.Add (rowNumberColumnName, rowNumberColumnName);
            rowNumber.Type = AttributeType.RowNumber;
            rowNumber.KeyUniquenessGuarantee = true;
            rowNumber.Usage = AttributeUsage.Key;
            rowNumber.KeyColumns.Add (new DataItem ());
            rowNumber.KeyColumns[0].DataType = System.Data.OleDb.OleDbType.Integer;
            rowNumber.KeyColumns[0].DataSize = 4;
            rowNumber.KeyColumns[0].NullProcessing = NullProcessing.Error;
            rowNumber.KeyColumns[0].Source = new RowNumberBinding ();
            rowNumber.NameColumn = new DataItem ();
            rowNumber.NameColumn.DataType = System.Data.OleDb.OleDbType.WChar;
            rowNumber.NameColumn.DataSize = 4;
            rowNumber.NameColumn.NullProcessing = NullProcessing.ZeroOrBlank;
            rowNumber.NameColumn.Source = new RowNumberBinding ();
            rowNumber.OrderBy = OrderBy.Key;
            rowNumber.AttributeHierarchyVisible = false;

            foreach (DataColumn dataColumn in dataTable.Columns) {
                DimensionAttribute newAttribute;
                string attributeID, attributeName;
                attributeID = dataColumn.ColumnName;
                attributeName = dataColumn.ColumnName;
                newAttribute = newDimension.Attributes.Add (attributeName, attributeID);
                newAttribute.Usage = AttributeUsage.Regular;
                newAttribute.KeyUniquenessGuarantee = false;
                newAttribute.KeyColumns.Add (new DataItem (dataTable.TableName, dataColumn.ColumnName, OleDbTypeConverter.GetRestrictedOleDbType (dataColumn.DataType)));
                newAttribute.KeyColumns[0].Source = new ColumnBinding (dataTable.TableName, dataColumn.ColumnName);
                newAttribute.KeyColumns[0].NullProcessing = NullProcessing.Preserve;
                newAttribute.NameColumn = new DataItem (dataTable.TableName, dataColumn.ColumnName, System.Data.OleDb.OleDbType.WChar);
                newAttribute.NameColumn.Source = new ColumnBinding (dataTable.TableName, dataColumn.ColumnName);
                newAttribute.NameColumn.NullProcessing = NullProcessing.ZeroOrBlank;
                newAttribute.OrderBy = OrderBy.Key;
                AttributeRelationship newAttributeRelationship = newDimension.Attributes[rowNumberColumnName].AttributeRelationships.Add (newAttribute.ID);
                newAttributeRelationship.Cardinality = Cardinality.Many;
                newAttributeRelationship.OverrideBehavior = OverrideBehavior.None;
            }
            modelCube.Dimensions.Add (tableName);

            //
            //  Create the measure group
            //
            MeasureGroup newMeasureGroup = modelCube.MeasureGroups.Add (tableName, tableName);
            newMeasureGroup.StorageMode = StorageMode.InMemory;
            newMeasureGroup.ProcessingMode = ProcessingMode.Regular;

            DegenerateMeasureGroupDimension newMeasureGroupDimension = new DegenerateMeasureGroupDimension (tableName);
            newMeasureGroup.Dimensions.Add (newMeasureGroupDimension);
            newMeasureGroupDimension.ShareDimensionStorage = StorageSharingMode.Shared;
            newMeasureGroupDimension.CubeDimensionID = newDimension.ID;

            foreach (DimensionAttribute ca in newDimension.Attributes) {
                MeasureGroupAttribute newMeasureGroupAttribute = new MeasureGroupAttribute (ca.ID);
                if (newMeasureGroupAttribute.AttributeID == rowNumberColumnName) {
                    newMeasureGroupAttribute.Type = MeasureGroupAttributeType.Granularity;
                    DataItem rowNumberKeyColumn = new DataItem (new ColumnBinding (tableName, rowNumberColumnName));
                    rowNumberKeyColumn.DataType = System.Data.OleDb.OleDbType.Integer;
                    newMeasureGroupAttribute.KeyColumns.Add (rowNumberKeyColumn);
                } else {
                    foreach (DataItem di in ca.KeyColumns) {
                        DataItem keyColumn = new DataItem (new ColumnBinding (tableName, ((ColumnBinding) di.Source).ColumnID));
                        keyColumn.DataType = di.DataType;
                        keyColumn.NullProcessing = NullProcessing.Preserve;
                        keyColumn.InvalidXmlCharacters = InvalidXmlCharacters.Remove;
                        newMeasureGroupAttribute.KeyColumns.Add (keyColumn);
                    }
                }
                newMeasureGroupDimension.Attributes.Add (newMeasureGroupAttribute);
            }

            //
            //  Adding default Measure to the measure group
            //
            String defaultMeasureID = string.Concat ("Count Of ", tableName);
            Measure currentMeasure = newMeasureGroup.Measures.Add (defaultMeasureID, defaultMeasureID);
            currentMeasure.AggregateFunction = AggregationFunction.Count;
            currentMeasure.DataType = MeasureDataType.BigInt;
            DataItem currentMeasureSource = new DataItem (new RowBinding (tableName));
            currentMeasureSource.DataType = System.Data.OleDb.OleDbType.BigInt;
            currentMeasure.Source = currentMeasureSource;

            //
            //  Adding default partition to the measure group
            //
            Partition currentPartition = new Partition (tableName, tableName);
            currentPartition.StorageMode = StorageMode.InMemory;
            currentPartition.ProcessingMode = ProcessingMode.Regular;
            currentPartition.Source = new QueryBinding (dataSourceView.DataSource.ID, tableQuery);
            newMeasureGroup.Partitions.Add (currentPartition);
        }

        /// <summary>
        /// 
        ///     Adds a measure to the database
        /// 
        /// </summary>
        /// <param name="Database"></param>
        /// <param name="tableName"></param>
        /// <param name="measureName"></param>
        /// <param name="measureExpression"></param>
        /// <param name="measureFormatString"></param>
        public void AddMeasure (string tableName, string measureName, string measureExpression, string measureFormatString) {
            Cube modelCube = Database.Cubes[0];
            MdxScript mdxScript = modelCube.MdxScripts["MdxScript"];

            //
            //  Beware that the comment lines should be there exactly as they are written, otherwise
            //  the SSDT editor will not recognize measures and it will not show them in the measure
            //  grid. See http://blogs.msdn.com/b/cathyk/archive/2011/11/29/powerpivot-measures-command-no-really-don-t-modify-it-manually.aspx
            //  for more information on this very stupid problem.
            //
            StringBuilder measuresCommand = new StringBuilder ();
            if (mdxScript.Commands.Count == 1) {
                measuresCommand.AppendLine ("-------------------------------------------------------------");
                measuresCommand.AppendLine ("-- Tabular Model measures command (do not modify manually) --");
                measuresCommand.AppendLine ("-------------------------------------------------------------");
                measuresCommand.AppendLine ();
                measuresCommand.AppendLine ();
                mdxScript.Commands.Add (new Command (measuresCommand.ToString ()));

            } else {
                measuresCommand.Append (mdxScript.Commands[1].Text);
            }
            measuresCommand.AppendLine (string.Format ("CREATE MEASURE '{0}'[{1}]={2};", tableName, measureName, measureExpression));
            mdxScript.Commands[1].Text = measuresCommand.ToString ();
            CalculationProperty newCalculationProperty = new CalculationProperty (measureName, CalculationType.Member);
            newCalculationProperty.FormatString = measureFormatString;
            newCalculationProperty.Visible = true;
            mdxScript.CalculationProperties.Add (newCalculationProperty);
            modelCube.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
        }

        /// <summary>
        /// 
        ///     Adds a calculated column to a table.
        ///     You need to create both the attribute in the dimension and the measure group attribute
        ///     in the dimension to make a calculated column work.
        /// 
        /// </summary>
        /// <param name="Database"></param>
        /// <param name="tableName"></param>
        /// <param name="calcColumnName"></param>
        /// <param name="calcColumnExpression"></param>
        public void AddCalculatedColumn (string tableName, string calcColumnName, string calcColumnExpression) {
            Dimension dim = Database.Dimensions[tableName];
            DimensionAttribute newDimensionAttribute = dim.Attributes.Add (calcColumnName);
            newDimensionAttribute.Usage = AttributeUsage.Regular;
            newDimensionAttribute.KeyUniquenessGuarantee = false;
            newDimensionAttribute.KeyColumns.Add (new DataItem (tableName, calcColumnName, OleDbType.Empty));
            newDimensionAttribute.KeyColumns[0].Source = new ExpressionBinding (calcColumnExpression);
            newDimensionAttribute.KeyColumns[0].NullProcessing = NullProcessing.Preserve;
            newDimensionAttribute.NameColumn = new DataItem (tableName, calcColumnName, System.Data.OleDb.OleDbType.WChar);
            newDimensionAttribute.NameColumn.Source = new ExpressionBinding (calcColumnExpression);
            newDimensionAttribute.NameColumn.NullProcessing = NullProcessing.ZeroOrBlank;
            newDimensionAttribute.OrderBy = OrderBy.Key;

            //
            //  Sets the attribute relationship with the table key
            //
            //  Corrected from Amo2Tabular: we no longer ser the relationship with RowNumber, we use the KeyAttribute instead,
            //  which seems a better choice in case a PK has been defined on the table
            //
            AttributeRelationship newAttributeRelationship = dim.KeyAttribute.AttributeRelationships.Add (newDimensionAttribute.ID);
            newAttributeRelationship.Cardinality = Cardinality.Many;
            newAttributeRelationship.OverrideBehavior = OverrideBehavior.None;

            Database.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
            dim.Process (ProcessType.ProcessFull);

            //
            //  Add CC as attribute to the MG
            //
            Cube modelCube = Database.Cubes[0];
            MeasureGroup measureGroup = modelCube.MeasureGroups[tableName];
            DegenerateMeasureGroupDimension measureGroupDimension = (DegenerateMeasureGroupDimension) measureGroup.Dimensions[tableName];
            MeasureGroupAttribute measureGroupAttribute = new MeasureGroupAttribute (calcColumnName);
            measureGroupAttribute.KeyColumns.Add (new DataItem (tableName, calcColumnName, OleDbType.Empty));
            measureGroupAttribute.KeyColumns[0].Source = new ExpressionBinding (calcColumnExpression);
            measureGroupDimension.Attributes.Add (measureGroupAttribute);

            //
            //  Update Dimension, CubeDimension and MG in server
            //
            Database.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
            measureGroup.Process (ProcessType.ProcessFull);
        }

        /// <summary>
        /// 
        ///     Creates a relationship between two tables
        /// 
        /// </summary>
        /// <param name="sourceTable"></param>
        /// <param name="sourceColumn"></param>
        /// <param name="targetTable"></param>
        /// <param name="targetColumn"></param>
        public void AddRelationship (string sourceTable, string sourceColumn, string targetTable, string targetColumn, bool makeActive) {
            Cube modelCube = Database.Cubes[0];
            //
            //  If the destination column is not the key of the table, we fix the table by setting it to be
            //  the key.
            //
            DimensionAttribute toAttribute = Database.Dimensions[targetTable].Attributes[targetColumn];
            if (toAttribute.Usage != AttributeUsage.Key) {
                //
                //  First step is to verify that the table has no keys, if it has some then we remove it and
                //  we set the relationship with the RowNumber column
                //
                DimensionAttribute rowNumber = Database.Dimensions[targetTable].Attributes[rowNumberColumnName];

                foreach (DimensionAttribute curAttr in Database.Dimensions[targetTable].Attributes) {
                    if (curAttr.ID != targetColumn && curAttr.Usage == AttributeUsage.Key) {
                        curAttr.Usage = AttributeUsage.Regular;
                        if (curAttr != rowNumber) {
                            curAttr.KeyColumns[0].NullProcessing = NullProcessing.Preserve;
                            curAttr.AttributeRelationships.Clear ();
                            if (!rowNumber.AttributeRelationships.Contains (targetColumn)) {
                                rowNumber.AttributeRelationships.Add (curAttr.ID);
                                rowNumber.AttributeRelationships[curAttr.ID].OverrideBehavior = OverrideBehavior.None;
                            }
                            rowNumber.AttributeRelationships[curAttr.ID].Cardinality = Cardinality.Many;
                        }
                    }
                }
                //
                //  Now the table has no key defined, we can set the desired column as the primary key of the
                //  table and remove its dependency from RowNumber. As it is the PK, we allow no NULL in the
                //  column content and we set a 1:1 attribute relationship with the RowNumber column
                //  
                rowNumber.AttributeRelationships.Remove (toAttribute.ID);
                toAttribute.Usage = AttributeUsage.Key;
                toAttribute.KeyColumns[0].NullProcessing = NullProcessing.Error;
                toAttribute.AttributeRelationships.Add (rowNumber.ID);
                toAttribute.AttributeRelationships[rowNumber.ID].OverrideBehavior = OverrideBehavior.None;
                toAttribute.AttributeRelationships[rowNumber.ID].Cardinality = Cardinality.One;
                Database.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
                Database.Dimensions[targetTable].Process (ProcessType.ProcessDefault);
            }

            //
            //  Now we are sure that the destination column is the PK of the table, we can proceed
            //  setting the relationship
            //
            string relName = String.Format ("Rel from {0}_{1} to {2}_{3}", sourceTable, sourceColumn, targetTable, targetColumn);
            Relationship newRelationship = Database.Dimensions[sourceTable].Relationships.Add (relName);
            newRelationship.FromRelationshipEnd.DimensionID = sourceTable;
            newRelationship.FromRelationshipEnd.Attributes.Add (sourceColumn);
            newRelationship.FromRelationshipEnd.Multiplicity = Multiplicity.Many;
            newRelationship.FromRelationshipEnd.Role = String.Empty;

            newRelationship.ToRelationshipEnd.DimensionID = targetTable;
            newRelationship.ToRelationshipEnd.Attributes.Add (targetColumn);
            newRelationship.ToRelationshipEnd.Multiplicity = Multiplicity.One;
            newRelationship.ToRelationshipEnd.Role = String.Empty;

            //
            //  Save the database and process the tables
            //
            Database.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
            Database.Dimensions[sourceTable].Process (ProcessType.ProcessDefault);
            Database.Dimensions[targetTable].Process (ProcessType.ProcessDefault);

            //
            //  Now it is time to activate the relationship
            //
            if (makeActive) {
                MeasureGroup fromMeasureGroup = modelCube.MeasureGroups[sourceTable];
                if (!fromMeasureGroup.Dimensions.Contains (targetTable)) {
                    MeasureGroupAttribute mgAttribute = new MeasureGroupAttribute (targetColumn);
                    mgAttribute.Type = MeasureGroupAttributeType.Granularity;
                    ReferenceMeasureGroupDimension newReferenceMGDim = new ReferenceMeasureGroupDimension ();
                    newReferenceMGDim.Attributes.Add (mgAttribute);
                    newReferenceMGDim.CubeDimensionID = targetTable;
                    newReferenceMGDim.IntermediateCubeDimensionID = sourceTable;
                    newReferenceMGDim.IntermediateGranularityAttributeID = sourceColumn;
                    newReferenceMGDim.Materialization = ReferenceDimensionMaterialization.Regular;
                    newReferenceMGDim.RelationshipID = relName;
                    fromMeasureGroup.Dimensions.Add (newReferenceMGDim);

                    ValidationErrorCollection errors = new ValidationErrorCollection ();
                    newReferenceMGDim.Validate (errors, true);
                    if (errors.Count > 0) {
                        StringBuilder errorMessages = new StringBuilder ();
                        foreach (ValidationError err in errors) {
                            errorMessages.AppendLine (string.Format ("At {2}: # {0} : {1}", err.ErrorCode, err.FullErrorText, err.Source));
                        }
                        throw new AmoException (errorMessages.ToString ());
                    }
                    //Update changes in the server
                    fromMeasureGroup.Update (UpdateOptions.ExpandFull, UpdateMode.CreateOrReplace);
                } else {
                    ReferenceMeasureGroupDimension toDimension = (ReferenceMeasureGroupDimension) fromMeasureGroup.Dimensions[targetTable];
                    toDimension.RelationshipID = relName;
                    toDimension.IntermediateGranularityAttributeID = sourceColumn;
                }
                fromMeasureGroup.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
                fromMeasureGroup.Process (ProcessType.ProcessFull);
            }
        }

        public void AddRelationship2 (string sourceTable, string sourceColumn, string targetTable, string targetColumn, bool makeActive) {
            Cube modelCube = Database.Cubes[0];
            //
            //  If the destination column is not the key of the table, we fix the table by setting it to be
            //  the key.
            //
            Dimension targetDimension = Database.Dimensions[targetTable];
            Dimension sourceDimension = Database.Dimensions[sourceTable];
            DimensionAttribute sourceAttribute = sourceDimension.Attributes[sourceColumn];
            DimensionAttribute targetAttribute = targetDimension.Attributes[targetColumn];
            DimensionAttribute rowNumberAttribute = Database.Dimensions[targetTable].Attributes[rowNumberColumnName];
            MeasureGroup sourceMeasureGroup = modelCube.MeasureGroups[sourceTable];

            rowNumberAttribute.AttributeRelationships[targetAttribute.ID].Cardinality = Cardinality.One;
            targetAttribute.KeyColumns[0].NullProcessing = NullProcessing.Error;

            //
            //  Now we are sure that the destination column is the PK of the table, we can proceed
            //  setting the relationship
            //
            string relName = String.Format ("Rel from {0}_{1} to {2}_{3}", sourceTable, sourceColumn, targetTable, targetColumn);
            Relationship newRelationship = sourceDimension.Relationships.Add (relName);
            newRelationship.FromRelationshipEnd.DimensionID = sourceDimension.ID;
            newRelationship.FromRelationshipEnd.Attributes.Add (sourceAttribute.ID);
            newRelationship.FromRelationshipEnd.Multiplicity = Multiplicity.Many;
            newRelationship.FromRelationshipEnd.Role = String.Empty;

            newRelationship.ToRelationshipEnd.DimensionID = targetDimension.ID;
            newRelationship.ToRelationshipEnd.Attributes.Add (targetAttribute.ID);
            newRelationship.ToRelationshipEnd.Multiplicity = Multiplicity.One;
            newRelationship.ToRelationshipEnd.Role = String.Empty;

            //
            //  Save the database and process the tables
            //
            Database.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
            Database.Dimensions[sourceTable].Process (ProcessType.ProcessDefault);
            Database.Dimensions[targetTable].Process (ProcessType.ProcessDefault);

            //
            //  Now it is time to activate the relationship
            //
            if (makeActive) {
                if (!sourceMeasureGroup.Dimensions.Contains (targetTable)) {
                    ReferenceMeasureGroupDimension newReferenceMGDim = new ReferenceMeasureGroupDimension ();
                    //
                    //  Adds all attributes from the target dimension to the current measure group 
                    //  setting the granularity to the key
                    //
                    foreach (DimensionAttribute da in targetDimension.Attributes) {
                        MeasureGroupAttribute mgAttribute = new MeasureGroupAttribute (da.Name);
                        if (targetAttribute.Name == da.Name) {
                            mgAttribute.Type = MeasureGroupAttributeType.Granularity;
                        }
                        newReferenceMGDim.Attributes.Add (mgAttribute);
                    }
                    newReferenceMGDim.CubeDimensionID = targetDimension.ID;
                    newReferenceMGDim.IntermediateCubeDimensionID = sourceDimension.ID;
                    newReferenceMGDim.IntermediateGranularityAttributeID = sourceAttribute.ID;
                    newReferenceMGDim.Materialization = ReferenceDimensionMaterialization.Regular;
                    newReferenceMGDim.RelationshipID = relName;
                    sourceMeasureGroup.Dimensions.Add (newReferenceMGDim);

                    List<Relationship> relationshipsToAdd = new List<Relationship> ();
                    List<Dimension> dimensionsToSearch = new List<Dimension> ();
                    dimensionsToSearch.Add (Database.Dimensions[targetTable]);
                    while (dimensionsToSearch.Count () > 0) {
                        foreach (Relationship r in dimensionsToSearch[0].Relationships) {
                            if (!sourceMeasureGroup.Dimensions.Contains (r.ToRelationshipEnd.DimensionID)) {
                                relationshipsToAdd.Add (r);
                                dimensionsToSearch.Add (Database.Dimensions[r.ToRelationshipEnd.DimensionID]);
                            }
                        }
                        dimensionsToSearch.RemoveAt (0);
                    }

                    foreach (Relationship r in relationshipsToAdd) {
                        //
                        //  Adds all attributes from the target dimension to the current measure group 
                        //  setting the granularity to the key
                        //
                        ReferenceMeasureGroupDimension rr = new ReferenceMeasureGroupDimension ();
                        foreach (DimensionAttribute da in Database.Dimensions[r.ToRelationshipEnd.DimensionID].Attributes) {
                            MeasureGroupAttribute mgAttribute = new MeasureGroupAttribute (da.Name);
                            if (r.ToRelationshipEnd.Attributes[0].AttributeID == da.Name) {
                                mgAttribute.Type = MeasureGroupAttributeType.Granularity;
                            }
                            rr.Attributes.Add (mgAttribute);
                        }
                        rr.CubeDimensionID = r.ToRelationshipEnd.DimensionID;
                        rr.IntermediateCubeDimensionID = r.FromRelationshipEnd.DimensionID;
                        rr.IntermediateGranularityAttributeID = r.ToRelationshipEnd.Attributes[0].AttributeID;
                        rr.Materialization = ReferenceDimensionMaterialization.Indirect;
                        rr.RelationshipID = r.ID;
                        sourceMeasureGroup.Dimensions.Add (rr);
                    }

                    ValidationErrorCollection errors = new ValidationErrorCollection ();
                    newReferenceMGDim.Validate (errors, true);
                    if (errors.Count > 0) {
                        StringBuilder errorMessages = new StringBuilder ();
                        foreach (ValidationError err in errors) {
                            errorMessages.AppendLine (string.Format ("At {2}: # {0} : {1}", err.ErrorCode, err.FullErrorText, err.Source));
                        }
                        throw new AmoException (errorMessages.ToString ());
                    }
                    //Update changes in the server
                    sourceMeasureGroup.Update (UpdateOptions.ExpandFull, UpdateMode.CreateOrReplace);
                } else {
                    ReferenceMeasureGroupDimension toDimension = (ReferenceMeasureGroupDimension) sourceMeasureGroup.Dimensions[targetTable];
                    toDimension.RelationshipID = relName;
                    toDimension.IntermediateGranularityAttributeID = sourceColumn;
                }
                sourceMeasureGroup.Update (UpdateOptions.ExpandFull, UpdateMode.UpdateOrCreate);
                sourceMeasureGroup.Process (ProcessType.ProcessFull);
            }
        }
    }
}
